from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Callable, Dict, Optional
from uuid import UUID, uuid4

from pydantic import BaseModel, Field

from agenticos.connectors.base import BaseWorkflowConfig

MSG_HS_NODE = "MSG_HS_NODE"
MSG_HS_ACK = "MSG_HS_ACK"
MSG_JOB_REQ = "MSG_JOB_REQ"
MSG_JOB_FIN = "MSG_JOB_FIN"
MSG_STEP_FIN = "MSG_STEP_FIN"
MSG_HEARTBEAT = "MSG_HEARTBEAT"

class Workflow(BaseModel):
    name: str
    description: str
    inputs: Dict[str, str]

class JobStatus(str, Enum):
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"


class AgenticConfig:
    def __init__(self, name: str, workflows: Dict[str, BaseWorkflowConfig]):
        self.name = name
        self.workflows = workflows

    id: Optional[UUID] = None
    name: str
    workflows: Dict[str, BaseWorkflowConfig] = {}

    def model_dump(self) -> dict:
        return {
            "id": self.id,
            "name": self.name,
            "workflows": {k: v.model_dump(k) for k, v in self.workflows.items()},
        }


class Job(BaseModel):
    id: UUID = Field(default_factory=uuid4)
    inputs: Dict[str, str]
    status: JobStatus = Field()
    output: str | None


class WrongFolderError(Exception):
    pass


class AgenticMessage(BaseModel):
    type: str


class AgenticHandshakeMessage(AgenticMessage):
    type: str = MSG_HS_NODE
    node: str


class JobFinishedMessage(AgenticMessage):
    type: str = MSG_JOB_FIN
    job_id: str
    status: JobStatus
    result: str | None


class StepFinishedMessage(AgenticMessage):
    type: str = MSG_STEP_FIN
    job_id: str
    step: int
    result: str


class JobRequest(BaseModel):
    workflow: str
    inputs: Dict[str, str]
    job_id: UUID = Field(default_factory=uuid4)
    node_id: str


class AgenticJobRequestMessage(BaseModel):
    type: str = MSG_JOB_REQ
    job: JobRequest



import importlib.util
import logging
import os
import sys
from subprocess import run

from dotenv import load_dotenv

from ..node.settings import HTTPSERVER, REGISTRY, settings
from ..node.wsnode import WSNode

log = logging.getLogger(__name__)


def run_agentic(*, mode: str, port: int, dev: bool, registry: str) -> None:
    if not os.path.isfile("pyproject.toml"):
        log.error("!!!!")
        log.error(
            "ERROR:You have to run this command in the root folder of your project"
        )
        log.error("!!!!")
        sys.exit(1)
    if mode == HTTPSERVER:
        cmd = ["fastapi"]
        cmd += ["dev", "--reload"] if dev else ["run"]
        cmd += ["--port", str(port)]
        cmd += ["src/agentic/agentic_node.py"]
        run(cmd, check=True)
    elif mode == REGISTRY:
        load_dotenv(verbose=True, dotenv_path=os.getcwd() + "/.env")
        # import workflows from the config file
        config_path = os.path.join(os.getcwd(), settings.AGENTIC_CONFIG_PATH)
        spec = importlib.util.spec_from_file_location("agentic_config", config_path)
        if spec is None:
            raise ValueError("Invalid config file")
        module = importlib.util.module_from_spec(spec)
        if spec.loader is None:
            raise ValueError("Invalid config file")
        src_path = settings.AGENTIC_CONFIG_PATH.split("/")[0]
        sys.path.append(os.getcwd() + "/" + src_path)
        spec.loader.exec_module(module)
        WSNode(config=module.config, registry=registry).connect_to_registry()

    else:
        raise ValueError("Invalid mode")



from abc import ABC, abstractmethod
from typing import Generator


class BaseWorkflowRunner(ABC):
    @abstractmethod
    def start(self, inputs : dict[str, str]) -> None:
        pass

    @abstractmethod
    def step_output(self) -> Generator[str, None, None]:
        pass


class BaseWorkflowConfig(ABC):
    @abstractmethod
    def description(self) -> str:
        pass

    @abstractmethod
    def inputs(self) -> dict[str, str]:
        pass

    @abstractmethod
    def steps_description(self) -> list[str]:
        pass

    @abstractmethod
    def new_runner(self) -> BaseWorkflowRunner:
        pass

    @abstractmethod
    def model_dump(self, name : str) -> dict:
        pass



from typing import Any, Generator, Type

from .base import BaseWorkflowConfig, BaseWorkflowRunner


class CrewaiWorkflowConfig(BaseWorkflowConfig):
    def __init__(self, description: str, inputs: dict[str, str], crew_cls: Type):
        self._description = description
        self._inputs = inputs
        crew_type = crew_cls.__name__
        self.crew_cls = crew_cls
        crew = crew_cls()
        if crew_type == "WrappedClass":
            self._steps_description = [t.description for t in crew.crew().tasks]
        elif crew_type == "Crew":
            self._steps_description = [t.description for t in crew.tasks]
        else:
            raise ValueError("Unknown crew type")

    def description(self) -> str:
        return self._description

    def inputs(self) -> dict[str, str]:
        return self._inputs

    def steps_description(self) -> list[str]:
        return self._steps_description

    def new_runner(self) -> BaseWorkflowRunner:
        return CrewaiWorkflowRunner(self.crew_cls)

    def model_dump(self, name: str) -> dict:
        return {
            "name": name,
            "description": self._description,
            "inputs": self._inputs,
            "steps_description": self._steps_description,
        }


class CrewaiWorkflowRunner(BaseWorkflowRunner):
    def __init__(self, crew_cls: Type):
        c = crew_cls()
        crew_type = crew_cls.__name__
        if crew_type == "WrappedClass":
            self.crew = c.crew()
        elif crew_type == "Crew":
            self.crew = c
        else:
            raise ValueError("Unknown crew type")

    def start(self, inputs: dict[str, str]) -> None:
        self._output = self.crew.kickoff(inputs=inputs)

    def step_output(self) -> Generator[str, None, None]:
        for task in self.crew.tasks:
            if hasattr(task, "output") and task.output:
                yield task.output.raw
